"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.waitTillBeenCalled = exports.itWithDedicatedCluster = exports.itWithCluster = exports.itWithClient = exports.describeHandleMinimumRedisVersion = exports.handleMinimumRedisVersion = exports.spawnGlobalRedisCluster = exports.spawnRedisCluster = exports.TEST_REDIS_CLUSTERES = exports.TestRedisClusters = exports.TEST_REDIS_SERVERS = exports.TestRedisServers = exports.isRedisVersionGreaterThan = void 0;
const assert_1 = require("assert");
const client_1 = __importDefault(require("./client"));
const child_process_1 = require("child_process");
const events_1 = require("events");
const which_1 = __importDefault(require("which"));
const cluster_1 = __importDefault(require("./cluster"));
const fs_1 = require("fs");
const utils_1 = require("./utils");
const REDIS_PATH = which_1.default.sync('redis-server'), REDIS_VERSION = getRedisVersion();
function getRedisVersion() {
    const raw = child_process_1.execSync(`${REDIS_PATH} -v`).toString(), indexOfVersion = raw.indexOf('v=');
    if (indexOfVersion === -1) {
        throw new Error('Unknown redis version');
    }
    const start = indexOfVersion + 2;
    return raw.substring(start, raw.indexOf(' ', start)).split('.', 3).map(Number);
}
function isRedisVersionGreaterThan(minimumVersion) {
    if (minimumVersion === undefined)
        return true;
    const lastIndex = minimumVersion.length - 1;
    for (let i = 0; i < lastIndex; i++) {
        if (REDIS_VERSION[i] > minimumVersion[i]) {
            return true;
        }
        else if (minimumVersion[i] > REDIS_VERSION[i]) {
            return false;
        }
    }
    return REDIS_VERSION[lastIndex] >= minimumVersion[lastIndex];
}
exports.isRedisVersionGreaterThan = isRedisVersionGreaterThan;
var TestRedisServers;
(function (TestRedisServers) {
    TestRedisServers[TestRedisServers["OPEN"] = 0] = "OPEN";
    TestRedisServers[TestRedisServers["PASSWORD"] = 1] = "PASSWORD";
})(TestRedisServers = exports.TestRedisServers || (exports.TestRedisServers = {}));
exports.TEST_REDIS_SERVERS = {};
var TestRedisClusters;
(function (TestRedisClusters) {
    TestRedisClusters[TestRedisClusters["OPEN"] = 0] = "OPEN";
})(TestRedisClusters = exports.TestRedisClusters || (exports.TestRedisClusters = {}));
exports.TEST_REDIS_CLUSTERES = {};
let port = 6379;
async function spawnRedisServer(args) {
    const currentPort = port++, process = child_process_1.spawn(REDIS_PATH, [
        '--save',
        '',
        '--port',
        currentPort.toString(),
        ...(args !== null && args !== void 0 ? args : [])
    ]);
    process
        .on('error', err => console.error('Redis process error', err))
        .on('close', code => console.error(`Redis process closed unexpectedly with code ${code}`));
    for await (const chunk of process.stdout) {
        if (chunk.toString().includes('Ready to accept connections')) {
            break;
        }
    }
    if (process.exitCode !== null) {
        throw new Error('Error while spawning redis server');
    }
    return {
        port: currentPort,
        async cleanup() {
            process.removeAllListeners('close');
            assert_1.strict.ok(process.kill());
            await events_1.once(process, 'close');
        }
    };
}
async function spawnGlobalRedisServer(args) {
    const { port, cleanup } = await spawnRedisServer(args);
    after(cleanup);
    return port;
}
const SLOTS = 16384;
async function spawnRedisClusterNode(type, nodeIndex, fromSlot, toSlot, args) {
    const clusterConfigFile = `/tmp/${type}-${nodeIndex}.conf`, { port, cleanup: originalCleanup } = await spawnRedisServer([
        '--cluster-enabled',
        'yes',
        '--cluster-node-timeout',
        '5000',
        '--cluster-config-file',
        clusterConfigFile,
        ...(args !== null && args !== void 0 ? args : [])
    ]);
    const client = client_1.default.create({
        socket: {
            port
        }
    });
    await client.connect();
    const range = [];
    for (let i = fromSlot; i < toSlot; i++) {
        range.push(i);
    }
    await Promise.all([
        client.clusterFlushSlots(),
        client.clusterAddSlots(range)
    ]);
    return {
        port,
        async cleanup() {
            await originalCleanup();
            try {
                await fs_1.promises.unlink(clusterConfigFile);
            }
            catch (err) {
                if (err.code == 'ENOENT')
                    return;
                throw err;
            }
        },
        client
    };
}
async function spawnRedisCluster(type, numberOfNodes, args) {
    const spawnPromises = [], slotsPerNode = Math.floor(SLOTS / numberOfNodes);
    for (let i = 0; i < numberOfNodes; i++) {
        const fromSlot = i * slotsPerNode;
        spawnPromises.push(spawnRedisClusterNode(type, i, fromSlot, i === numberOfNodes - 1 ? SLOTS : fromSlot + slotsPerNode, args));
    }
    const spawnResults = await Promise.all(spawnPromises), meetPromises = [];
    for (let i = 1; i < spawnResults.length; i++) {
        meetPromises.push(spawnResults[i].client.clusterMeet('127.0.0.1', spawnResults[i - 1].port));
    }
    await Promise.all(meetPromises);
    while ((await spawnResults[0].client.clusterInfo()).state !== 'ok') {
        await utils_1.promiseTimeout(100);
    }
    const disconnectPromises = [];
    for (const result of spawnResults) {
        disconnectPromises.push(result.client.disconnect());
    }
    await Promise.all(disconnectPromises);
    return spawnResults;
}
exports.spawnRedisCluster = spawnRedisCluster;
async function spawnGlobalRedisCluster(type, numberOfNodes, args) {
    const results = await spawnRedisCluster(type, numberOfNodes, args);
    after(() => Promise.all(results.map(({ cleanup }) => cleanup())));
    return results.map(({ port }) => port);
}
exports.spawnGlobalRedisCluster = spawnGlobalRedisCluster;
async function spawnOpenServer() {
    exports.TEST_REDIS_SERVERS[TestRedisServers.OPEN] = {
        port: await spawnGlobalRedisServer()
    };
}
async function spawnPasswordServer() {
    exports.TEST_REDIS_SERVERS[TestRedisServers.PASSWORD] = {
        port: await spawnGlobalRedisServer(['--requirepass', 'password']),
        password: 'password'
    };
    if (isRedisVersionGreaterThan([6])) {
        exports.TEST_REDIS_SERVERS[TestRedisServers.PASSWORD].username = 'default';
    }
}
async function spawnOpenCluster() {
    exports.TEST_REDIS_CLUSTERES[TestRedisClusters.OPEN] = (await spawnGlobalRedisCluster(TestRedisClusters.OPEN, 3)).map(port => ({
        port
    }));
}
before(function () {
    this.timeout(10000);
    return Promise.all([
        spawnOpenServer(),
        spawnPasswordServer(),
        spawnOpenCluster()
    ]);
});
function handleMinimumRedisVersion(mochaContext, minimumVersion) {
    if (isRedisVersionGreaterThan(minimumVersion)) {
        return false;
    }
    mochaContext.skip();
    return true;
}
exports.handleMinimumRedisVersion = handleMinimumRedisVersion;
function describeHandleMinimumRedisVersion(minimumVersion) {
    before(function () {
        handleMinimumRedisVersion(this, minimumVersion);
    });
}
exports.describeHandleMinimumRedisVersion = describeHandleMinimumRedisVersion;
function itWithClient(type, title, fn, options) {
    it(title, async function () {
        if (handleMinimumRedisVersion(this, options === null || options === void 0 ? void 0 : options.minimumRedisVersion))
            return;
        const client = client_1.default.create({
            socket: exports.TEST_REDIS_SERVERS[type]
        });
        await client.connect();
        try {
            await client.flushAll();
            await fn(client);
        }
        finally {
            await client.flushAll();
            await client.disconnect();
        }
    });
}
exports.itWithClient = itWithClient;
function itWithCluster(type, title, fn, options) {
    it(title, async function () {
        if (handleMinimumRedisVersion(this, options === null || options === void 0 ? void 0 : options.minimumRedisVersion))
            return;
        const cluster = cluster_1.default.create({
            rootNodes: exports.TEST_REDIS_CLUSTERES[type]
        });
        await cluster.connect();
        try {
            await clusterFlushAll(cluster);
            await fn(cluster);
        }
        finally {
            await clusterFlushAll(cluster);
            await cluster.disconnect();
        }
    });
}
exports.itWithCluster = itWithCluster;
function itWithDedicatedCluster(title, fn) {
    it(title, async function () {
        this.timeout(10000);
        const spawnResults = await spawnRedisCluster(null, 3), cluster = cluster_1.default.create({
            rootNodes: [{
                    port: spawnResults[0].port
                }]
        });
        await cluster.connect();
        try {
            await fn(cluster);
        }
        finally {
            await cluster.disconnect();
            for (const { cleanup } of spawnResults) {
                await cleanup();
            }
        }
    });
}
exports.itWithDedicatedCluster = itWithDedicatedCluster;
async function clusterFlushAll(cluster) {
    await Promise.all(cluster.getMasters().map(({ client }) => client.flushAll()));
}
async function waitTillBeenCalled(spy) {
    const start = process.hrtime.bigint(), calls = spy.callCount;
    do {
        if (process.hrtime.bigint() - start > 1000000000) {
            throw new Error('Waiting for more than 1 second');
        }
        await utils_1.promiseTimeout(1);
    } while (spy.callCount === calls);
}
exports.waitTillBeenCalled = waitTillBeenCalled;
